/*
 *  @(#) dlfcn.h 1.0
 *
 *  (c) Copyright 1996 Hewlett-Packard Co., All Rights Reserved.
 */

#ifndef _DLFCN_INCLUDED
#define _DLFCN_INCLUDED
#include <sys/types.h>
#ifdef __cplusplus
extern "C" {
#endif

#ifdef __LP64__
typedef unsigned long		UINT64;
#else
typedef unsigned long long	UINT64;
#endif

/* Current libdl.sl version */
#define DL_CURRENT_VERSION	1

/* A Load module descriptor structure */
#ifndef _LOAD_MODULE_DESC_T
#define _LOAD_MODULE_DESC_T
struct load_module_desc {
   UINT64	text_base;
   UINT64	text_size;
   UINT64	data_base;
   UINT64	data_size;
   UINT64	unwind_base;
   UINT64	linkage_ptr;
   UINT64	phdr_base;
   UINT64       tls_size;
   UINT64       tls_start_addr;
   UINT64       unwind_size;
#ifdef NUMA
   UINT64       ns_size;
   UINT64       ns_base;
   UINT64       fs_size;
   UINT64       fs_base;
#endif /* NUMA */
};
#endif /* ! _LOAD_MODULE_DESC_T */

struct dld_hook_param {
   unsigned int version;
   char* filename; 
   union {
      struct { 
         unsigned int startup_load; 
         struct load_module_desc* desc; 
         size_t desc_size; 
      } load_event_desc; 
      struct { 
         struct load_module_desc* desc; 
         size_t desc_size; 
      } unload_event_desc; 
      struct { 
	 char*  sym_name;
         struct load_module_desc* desc; 
         size_t desc_size; 
      } bor_event_desc; 
   } event_desc;
};

#define load_event 	event_desc.load_event_desc
#define unload_event 	event_desc.unload_event_desc
#define bor_event 	event_desc.bor_event_desc

#pragma HP_DEFINED_EXTERNAL	dlopen
#pragma HP_DEFINED_EXTERNAL	dlsym 
#pragma HP_DEFINED_EXTERNAL	dlclose
#pragma HP_DEFINED_EXTERNAL	dlget
#pragma HP_DEFINED_EXTERNAL	dlmodinfo
#pragma HP_DEFINED_EXTERNAL	dlgetmodinfo
#pragma HP_DEFINED_EXTERNAL	dlgetname
#pragma HP_DEFINED_EXTERNAL	dlhook
#pragma HP_DEFINED_EXTERNAL	dlerror

#if defined(__STDC__) || defined(__cplusplus)
extern void*	dlopen(const char* filename,
                       unsigned int flags);
extern void*	dlsym(void* handle,
	              const char* name);
extern int	dlclose(void* handle);
extern void*	dlget(unsigned int index, 
                      struct load_module_desc* desc,
                      unsigned int desc_size);
extern UINT64	dlmodinfo(UINT64 ip_value,
                          struct load_module_desc* desc,
                          size_t desc_size,
                          void* (*read_tgt_mem)(void* buffer,
                                                UINT64 ptr,
                                                size_t bufsiz,
                                                int ident),
                          int ident_parm,
                          UINT64 load_map_parm);
extern UINT64	dlgetmodinfo(int index,
			     struct load_module_desc* desc,
			     size_t desc_size,
			     void* (*read_tgt_mem)(void* buffer,
						   UINT64 ptr,
						   size_t bufsiz,
						   int ident),
			     int ident_parm,
			     UINT64 load_map_parm);
extern char*	dlgetname(struct load_module_desc* desc,
                          size_t desc_size,
                          void* (*read_tgt_mem)(void* buffer,
                                                UINT64 ptr,
                                                size_t bufsiz,
                                                int ident),
                          int ident_parm,
                          UINT64 load_map_parm);
extern int	dlhook(unsigned int flags, 
		       void (*dld_hook)(unsigned int type,
					struct dld_hook_param* arg));
extern char*	dlerror(void);
#else /* not __STDC__ || __cplusplus */
extern void*	dlopen();
extern void*	dlsym();
extern int	dlclose();
extern void*	dlget();
extern UINT64	dlmodinfo();
extern UINT64	dlgetmodinfo();
extern char*	dlgetname();
extern int	dlhook();
extern char*	dlerror();
#endif /* not __STDC__ || __cplusplus */

/* dlopen flags */
#define RTLD_NOW        0x00000001  /* bind immediately */
#define RTLD_LAZY       0x00000002  /* bind deferred */
#define RTLD_GLOBAL     0x00000004  /* symbols are globally visible */
#define RTLD_LOCAL      0x00000008  /* symbols only visible to this load */

/* dlsym special next handle */
#define RTLD_NEXT	((void*) -1L)  /* start symbol search from the dll
					  next to the caller of dlsym() */
/* dlhook event flags */
#define DL_LOAD_PRE_INIT	0x00000001  /* load event before initializers */
#define DL_LOAD_POST_INIT	0x00000002  /* load event after initializers */
#define DL_LOAD_COMPLETE	0x00000004  /* load event complete */
#define DL_UNLOAD_PRE_FINI	0x00000008  /* unload event before terminators*/
#define DL_UNLOAD_POST_FINI	0x00000010  /* unload event after terminators */
#define DL_UNLOAD_COMPLETE	0x00000020  /* unload event complete */
#define DL_BOR			0x00000040  /* bind-on-reference event */

#ifdef __cplusplus
}
#endif
#endif /* _DLFCN_INCLUDED */
