/*
	GLfunge98
	OpenGL-based implementation of Funge98
	Copyright (C) 2000, Heikki Kallasjoki

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA	
*/


#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#ifndef WIN32
#ifndef __hpux
#include <dlfcn.h>
#endif
#endif /* !WIN32 */
#ifdef OUT_GL
#include <out_GL.h>
#endif /* OUT_GL */
#ifdef OUT_TEXT
#include <out_text.h>
#endif /* OUT_TEXT */
extern char **environ;


/* required structures, like the IP, and global variables */
#include <IPstruct.h>
unsigned long int IPcount;
unsigned long int IPstacksize;
IP ** IPs;
/* we make the space from 64x64 (16 kilobytes) blocks allocated as needed */
typedef struct {
	long int offsetx, offsety;
	long int data[64][64];
} spaceblock;
unsigned long int SBcount;
unsigned long int SBstacksize;
spaceblock ** SBs;
spaceblock * sbquick[16][16];
long int blockmaxx, blockmaxy, blockminx, blockminy;
long int lastgetx, lastgety, lastgetv;
int cmdargc; char ** cmdargv;
int pf_minx = 0; int pf_maxx = 79; int pf_miny = 0; int pf_maxy = 24;
#define MODE_FX 0
#define MODE_RUN 1
#define MODE_DEBUG 2
#define MODE_EDIT 3
int progmode;
#define OUTMODE_TEXT 0
#define OUTMODE_GL 1
#ifdef OUT_GL
int outputmode = OUTMODE_GL;
extern unsigned long int framecount_GL;
#else
int outputmode = OUTMODE_TEXT;
#endif /* OUT_GL */
#define FLAG_BEFUNGE93  0x00000001
#define FLAG_DEBUG      0x00000002
#define FLAG_LOGALL     0x00000004
#define FLAG_SINGLESTEP 0x00000008
#define FLAG_STAT       0x00000010
#define FLAG_LIMITED    0x00000020
unsigned long int flags;
struct s_statis {
	int max_IP;
	int max_stack;
	int ticks;
	int minstacknum;
	int maxstacknum;
	struct timeval starttime;
} statinfo;
FILE * logfile = NULL;
char * timestr = NULL;

const char version_str[] = "0.9.04";
const long int version_num = 904;

/* initworld initializes our world, the befunge-space and IP stack and stuff */
void initworld(void);
/* createIP creates a new IP, with its own stack etc. It can use and old IP to create a clone of it (for the 't' instruction) */
IP * createIP(IP *);
/* destroyIP destroys an IP, and terminates the program if necessary */
void destroyIP(IP *);
/* execIP executes the command from the place the IP is */
int execIP(IP *, long int);
/* moveIP moves the IP according to delta, and warps spaces. moveIP93 is the befunge93-compatible version */
void moveIP(IP *);
void moveIP93(IP *);
/* pop pops a signed integer from the stack of IP. push pushes one. */
long int pop(IP *);
void push(IP *, long int);
/* put puts a value to the funge-space, get gets one */
long int get(long int, long int);
void put(long int, long int, long int);
/* systeminfo retrieves lots of information about the system */
void systeminfo(IP *);
/* logtime and logclose do things for logging system */
char * logtime(void);
void logclose(void);
/* loadext tries to load an extension from GLfunge98_ext_[fingerprint].so (in *nix) or GLfunge98_ext_[fingerprint].dll (in w32) */
/*
int loadext(IP *, long int);
int unloadext(IP *);
*/

/* temporarily, for clean compilation: */
int outfile(IP * curIP) { return 0; }
int inpfile(IP * curIP) { return 0; }
void fail(char * x) {
	fprintf(stderr, "%s\n", x);
	exit(1);
}
void terminate(int x) {
	int i; struct timeval nowtime; float dif;

	if(flags & FLAG_STAT) gettimeofday(&nowtime, NULL);

#ifdef OUT_GL
	if(outputmode == OUTMODE_GL) uninit_GL();
#endif /* OUT_GL */

	if(flags & FLAG_STAT) {
		dif = nowtime.tv_sec - statinfo.starttime.tv_sec;
		dif += ((float)(nowtime.tv_usec - statinfo.starttime.tv_usec)) / 1000000;
		printf("\n-- Statistics: --\n");
		printf(" - Maximum number of IPs: %d\n", statinfo.max_IP);
		printf(" - Maximum size of stack: %d\n", statinfo.max_stack);
		printf(" - Smallest and largest number in stack: %d, %d\n", statinfo.minstacknum, statinfo.maxstacknum);
		printf(" - Funge-clock ticks used: %d\n", statinfo.ticks);
		printf(" - Real world time used: %.3f s\n", dif);
		if(statinfo.ticks / dif > 1000000)
			printf(" - MFIPS: %.3f\n", (statinfo.ticks / dif) / 1000000);
		else
			printf(" - KFIPS: %.3f\n", (statinfo.ticks / dif) / 1000);
	}
	
	/*for(i = 0; i < IPcount; i++) while(unloadext(IPs[i]));*/

	exit(x);
}
long int execcmd(IP * curIP) { return 0; }


int main(int argc, char ** argv) {
	int i, j, k, l;
	char * fname_log, * fname_prog, * tmpstr;
	FILE * progfile = NULL;
#ifdef OUT_GL
	struct itimerval myItimerval;
#endif /* OUT_GL */

	printf("GLfunge98-%s.\n", version_str);
	
	/* save command arguments for Funge98-'y'-command */
	cmdargc = argc;
	cmdargv = argv;

	/* defaults, for now. later from .GLfunge98rc */
#ifdef OUT_GL
	progmode = MODE_FX;
#endif /* OUT_GL */
#ifdef OUT_TEXT
	progmode = MODE_RUN;
#endif /* OUT_TEXT */
	flags = 0;
	fname_log = NULL; fname_prog = NULL;
	
	/* and command line argument processing */
	for(i = 1; i < argc; i++) {
		if((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-?") == 0) || (strcmp(argv[i], "--help") == 0)) {
			printf("Usage: %s [options] [filename]\n", argv[0]);
			printf(" filename is the name of a Funge-program source file.\n");
			printf(" Extensions .glf (GLfunge98-file), .b98 and .bf will be tried if not found.\n");
			printf(" Options:\n");
			printf("  Output modes:\n");
#ifdef OUT_TEXT
			printf("  -t or --text: run in text/console mode. %s\n", (outputmode == OUTMODE_TEXT ? "(default)" : ""));
#endif /* OUT_TEXT */
#ifdef OUT_GL
			printf("  -g or --opengl: run in OpenGL mode. %s\n", (outputmode == OUTMODE_GL ? "(default)" : ""));
#endif /* OUT_GL */
			printf("  Modes:\n");
			printf("  -r or --run: execute with 'run' mode, only program output will be shown. %s\n", (outputmode == OUTMODE_TEXT ? "(default)" : ""));
			printf("  -d or --debug: execute with 'debug' mode.\n");
#ifdef OUT_GL
			printf("  -f or --fx: execute with 'fx' mode, with k3wl graphical effects %s.\n", (outputmode == OUTMODE_GL ? "(default)" : ""));
#endif /* OUT_GL */
			printf("  -e or --edit: start GLfunge98 in editor mode, to develop new programs.\n");
			printf("  Misc. options:\n");
			printf("  -93 or --befunge93: run in Befunge93-compatibility mode.\n");
			printf("  -w[x] or --width [x]: restricts playfield width to x columns.\n");
			printf("  -h[x] or --height [x]: restricts playfield height to x rows.\n");
			printf("  -l[file] or --log [file]: log output of the program to file.\n");
			printf("  -L[file] or --logall [file]: log every instruction, stack changes, etc.\n");
			printf("  -s or --statistics: print statistics after program has exited.\n");
			printf("  -V or --version: print GLfunge98 version number and exit.\n");
			printf("  -D: print lots of unnecessary debug info about the interpreter.\n");
			exit(0);
			continue;
		}
#ifdef OUT_TEXT
		if((strcmp(argv[i], "-t") == 0) || (strcmp(argv[i], "--text") == 0)) {
			outputmode = OUTMODE_TEXT;
			continue;
		}
#endif /* OUT_TEXT */
#ifdef OUT_GL
		if((strcmp(argv[i], "-g") == 0) || (strcmp(argv[i], "--opengl") == 0)) {
			outputmode = OUTMODE_GL;
			continue;
		}
#endif /* OUT_GL */
		if((strcmp(argv[i], "-r") == 0) || (strcmp(argv[i], "--run") == 0)) {
			progmode = MODE_RUN;
			continue;
		}
		if((strcmp(argv[i], "-d") == 0) || (strcmp(argv[i], "--debug") == 0)) {
			progmode = MODE_DEBUG;
			/* flags |= FLAG_SINGLESTEP; */
			continue;
		}
#ifdef OUT_GL
		if((strcmp(argv[i], "-f") == 0) || (strcmp(argv[i], "--fx") == 0)) {
			progmode = MODE_FX;
			continue;
		}
#endif /* OUT_GL */
		if((strcmp(argv[i], "-e") == 0) || (strcmp(argv[i], "--edit") == 0)) {
			progmode = MODE_EDIT;
			continue;
		}
		if((strcmp(argv[i], "-93") == 0) || (strcmp(argv[i], "--befunge93") == 0)) {
			flags |= FLAG_BEFUNGE93 | FLAG_LIMITED;
			continue;
		}
		if(strncmp(argv[i], "-w", 2) == 0) {
			tmpstr = argv[i] + 2;
			pf_maxx = atoi(tmpstr) - 1;
			flags |= FLAG_LIMITED;
			continue;
		}
		if(strcmp(argv[i], "--width") == 0) {
			tmpstr = argv[++i];
			pf_maxx = atoi(tmpstr) - 1;
			flags |= FLAG_LIMITED;
			continue;
		}
		if(strncmp(argv[i], "-h", 2) == 0) {
			tmpstr = argv[i] + 2;
			pf_maxy = atoi(tmpstr) - 1;
			flags |= FLAG_LIMITED;
			continue;
		}
		if(strcmp(argv[i], "--height") == 0) {
			tmpstr = argv[++i];
			pf_maxy = atoi(tmpstr) - 1;
			flags |= FLAG_LIMITED;
			continue;
		}
		if(strncmp(argv[i], "-l", 2) == 0) {
			fname_log = argv[i] + 2;
			continue;
		}
		if(strcmp(argv[i], "--log") == 0) {
			fname_log = argv[++i];
			continue;
		}
		if(strncmp(argv[i], "-L", 2) == 0) {
			fname_log = argv[i] + 2;
			flags |= FLAG_LOGALL;
			continue;
		}
		if(strcmp(argv[i], "--logall") == 0) {
			fname_log = argv[++i];
			flags |= FLAG_LOGALL;
			continue;
		}
		if((strcmp(argv[i], "-s") == 0) || (strcmp(argv[i], "--statistics") == 0)) {
			flags |= FLAG_STAT;
			continue;
		}
		if((strcmp(argv[i], "-V") == 0) || (strcmp(argv[i], "--version") == 0)) {
			exit(0);
		}
		if(strcmp(argv[i], "-D") == 0) {
			flags |= FLAG_DEBUG;
			continue;
		}
		if(!fname_prog) fname_prog = argv[i];
	}

	if((pf_maxx < 0) || (pf_maxy < 0)) {
		fprintf(stderr, "Invalid values for playfield size.\n");
		return 0;
	}
	if(!fname_prog) {
		printf("No program file given to execute.\nUsage instructions with \"%s -h\".\n", argv[0]);
		exit(0);
	}

	/* try some extensions */
	i = 0;
	tmpstr = (char *)malloc(strlen(fname_prog) + 5);
	strcpy(tmpstr, fname_prog);
	while(!progfile) {
		if(i == 1) strcpy(tmpstr + strlen(fname_prog), ".glf");
		if(i == 2) strcpy(tmpstr + strlen(fname_prog), ".b98");
		if(i == 3) strcpy(tmpstr + strlen(fname_prog), ".bf");
		if(i == 4) {
			printf("Unable to open file \"%s\".\n", fname_prog);
			exit(0);
		}
		progfile = fopen(tmpstr, "r");
		i++;
	}
	free(tmpstr);

	/* log file setup */
	if(fname_log) {
		logfile = fopen(fname_log, "a");
		if(!logfile) {
			printf("Unable to open log file \"%s\".\n", fname_log);
			exit(0);
		}
		fprintf(logfile, "[%s] --- GLfunge98-%s log begins ---\n", logtime(), version_str);
		fprintf(logfile, "Program file: %s\n", fname_prog);
		fprintf(logfile, "Interpreter options:");
		if(progmode == MODE_FX) fprintf(logfile, " [FX mode]");
		if(progmode == MODE_RUN) fprintf(logfile, " [run mode]");
		if(progmode == MODE_DEBUG) fprintf(logfile, " [debug mode]");
		if(flags & FLAG_BEFUNGE93) fprintf(logfile, " [Befunge93 compatibility]");
		if(flags & FLAG_DEBUG) fprintf(logfile, " [interpreter debug mode]");
		if(flags & FLAG_LOGALL) fprintf(logfile, " [logging all details]");
		fprintf(logfile, "\n");
	}

	/* initialize our world */
	initworld();

	/* load program to fungespace */
	tmpstr = (char *)malloc(1024);
	if(!tmpstr) {
		printf("Not enough memory to load program.\n");
		if(logfile) fprintf(logfile, "Not enough memory to load program.\n");
		logclose();
		if(flags & FLAG_DEBUG) fprintf(stderr, "malloc() failed for temporary string.\n");
		exit(0);
	}
	
	if(logfile) fprintf(logfile, "Loading program to Funge-space.\n");
	
	j = 0; k = 0; l = 0;
	while(fgets(tmpstr, 1024, progfile)) {
		for(i = 0; i < 1024; i++) {
			if(tmpstr[i] == '\0') break;
			if(tmpstr[i] == '\n') { j++; l = 1; k = 0; continue; }
			if(tmpstr[i] == '\r') {
				if(tmpstr[i + 1] == '\n') i++;
				j++; l = 1; k = 0; continue;
			}
			put(k + i, j, tmpstr[i]);
		}
		if(!l) k += i;
		l = 0;
	}

	if(logfile) fprintf(logfile, "Program loaded to Funge-space, %d lines.\n", j);

	/* install signal handler for program termination */
	signal(SIGINT, terminate);

	/* Main loop for code execution */
#ifdef OUT_GL
	if(outputmode == OUTMODE_GL) {
		/* register the timer */
		myItimerval.it_interval.tv_sec = 0;
		myItimerval.it_interval.tv_usec = 200000;
		myItimerval.it_value.tv_sec = 0;
		myItimerval.it_value.tv_usec = 200000;
		setitimer(ITIMER_VIRTUAL, &myItimerval, NULL);

		for(;;) {
			j = IPcount;
			for(i = 0; i < j; i++) {
				k = execIP(IPs[i], 0);
				if(!k) {
					if(i > 0) i--;
					else i = j - 1;
				} else {
					if(flags & FLAG_BEFUNGE93) moveIP93(IPs[i]);
					else moveIP(IPs[i]);
				}
				if(flags & FLAG_SINGLESTEP) inpchr_GL();
			}
			if(framecount_GL) { framecount_GL = 0; draw_GL(); }
		}
	}
#endif /* OUT_GL */
#ifdef OUT_TEXT
	if(outputmode == OUTMODE_TEXT) {
		for(;;) {
			j = IPcount;
			for(i = 0; i < j; i++) {
				k = execIP(IPs[i], 0);
				if(!k) {
					if(i > 0) i--;
					else i = j - 1;
				} else {
					if(flags & FLAG_BEFUNGE93) moveIP93(IPs[i]);
					else moveIP(IPs[i]);
				}
				if(flags & FLAG_SINGLESTEP) inpchr();
			}
		}
	}
#endif /* OUT_TEXT */

	return 0;
}

void initworld(void) {
	spaceblock * firstblock; long int k, l;

	/* some sensible values for global variables */
	if(flags & FLAG_STAT) {
		statinfo.max_IP = 0; statinfo.max_stack = 0;
		statinfo.maxstacknum = 0; statinfo.minstacknum = 0;
		statinfo.ticks = 0;
		gettimeofday(&statinfo.starttime, NULL);
	}
	for(k = 0; k < 16; k++) for(l = 0; l < 16; l++) sbquick[k][l] = NULL;

	/* initialize IP stack, create the first IP */
	IPs = (IP **)malloc(sizeof(IP *) * 256);
	if(!IPs) fail("unable to initialize IP stack");
	IPcount = 0;
	IPstacksize = 256;
	if(!createIP(NULL)) fail("unable to create the primary IP");

	/* create the first block */
	blockmaxx = 0; blockmaxy = 0; blockminx = 0; blockminy = 0;
	firstblock = (spaceblock *)malloc(sizeof(spaceblock));
	if(!firstblock) fail("unable to create the fungespace");
	SBcount = 1;
	SBstacksize = 256;
	SBs = (spaceblock **)malloc(256);
	if(!SBs) fail("unable to create fungespace block stack");
	SBs[0] = firstblock;
	firstblock->offsetx = 0; firstblock->offsety = 0;
	for(k = 0; k < 4096; k++) ((long int *)firstblock->data)[k] = 32;

#ifdef OUT_GL
	if(outputmode == OUTMODE_GL)
		if((k = init_GL())) {
			char errstr[128];
			sprintf(errstr, "OpenGL requested but unable to initialize it (%ld)", k);
			fail(errstr);
		}
#endif /* OUT_GL */
	
	return;
}

IP * createIP(IP * oldIP) {
	IP * newIP; IP ** tmpIPs;

	/* reserve space for a new IP structure */
	newIP = (IP *)malloc(sizeof(IP));
	if(!newIP) return NULL;
	
	if(!oldIP) {
		/* create a new stack and initialize default directions etc. */
		newIP->stack = (long int *)malloc(sizeof(long int) * 1024);
		if(!newIP->stack) { free(newIP); return NULL; }
		newIP->stacksize = 1024;
		newIP->SP = 0;
		
		newIP->posx = 0; newIP->posy = 0;
		newIP->dx = 1; newIP->dy = 0;
		newIP->soffx = 0; newIP->soffy = 0;
		newIP->stringmode = 0;
		newIP->curlib = NULL; memset(newIP->extfunctable, 0, sizeof(void *) * 26);
		newIP->getfunc = get; newIP->putfunc = put; newIP->popfunc = pop; newIP->pushfunc = push;
#ifdef OUT_GL
		if(outputmode == OUTMODE_GL) {
			newIP->outchrfunc = outchr_GL; newIP->outdecfunc = outdec_GL; newIP->inpchrfunc = inpchr_GL; newIP->inpdecfunc = inpdec_GL;
		}
#endif /* OUT_GL */
#ifdef OUT_TEXT
		if(outputmode == OUTMODE_TEXT) {
			newIP->outchrfunc = outchr; newIP->outdecfunc = outdec; newIP->inpchrfunc = inpchr; newIP->inpdecfunc = inpdec;
		}
#endif /* OUT_TEXT */

		newIP->prevlib = NULL; newIP->flags = &flags;
	} else {
		/* copy the old IP structure, then create a new stack and duplicate the old */
		memcpy(newIP, oldIP, sizeof(IP));
		newIP->stack = (long int *)malloc(sizeof(long int) * newIP->stacksize);
		if(!newIP->stack) { free(newIP); return NULL; }
		memcpy(newIP->stack, oldIP->stack, newIP->stacksize);
	}

	/* increase IP count, give the IP its number, enlarge the IP stack if necessary */
	IPcount++;
	if(flags & FLAG_STAT) if(statinfo.max_IP < IPcount) statinfo.max_IP++;
	newIP->num = IPcount - 1;
	if(IPcount >= IPstacksize) {
		IPstacksize += 256;
		tmpIPs = (IP **)malloc(sizeof(IP *) * IPstacksize);
		if(!tmpIPs) fail("malloc() failed for a new IP stack");
		memcpy(tmpIPs, IPs, sizeof(IP *) * (IPstacksize - 256));
		free(IPs);
		IPs = tmpIPs;
	}
	IPs[newIP->num] = newIP;

	return newIP;
}

void destroyIP(IP * oldIP) {
	unsigned long int i;

	IPcount--;
	if(!IPcount) terminate(0);
	
	if(oldIP->num != IPcount) memmove(IPs + oldIP->num, IPs + oldIP->num + 1, sizeof(IP *) * (IPcount - oldIP->num));
	for(i = oldIP->num; i < IPcount; i++) IPs[i]->num = i;

	free(oldIP->stack);
	free(oldIP);
	
	return;
}

int execIP(IP * curIP, long int cmd) {
	long int i, a, b, c, r; IP * newIP;
	int (*extfunc)(IP *);

	r = 1;

	if(curIP == NULL) c = cmd;
	else c = get(curIP->posx, curIP->posy);

	if(flags & FLAG_STAT) statinfo.ticks++;

	/* special cases */
	if(curIP->stringmode && (c != '\"')) {
		if(logfile && (flags & FLAG_LOGALL)) fprintf(logfile, "execIP(IP:%ld): stringmode: char '%c'.\n", curIP->num, (int)c);
		if(progmode == MODE_DEBUG) fprintf(stderr, "execIP(IP:%ld): stringmode: char '%c'.\n", curIP->num, (int)c);
		push(curIP, c); return r;
	}

	if(logfile && (flags & FLAG_LOGALL)) {
		fprintf(logfile, "execIP(IP:%ld): pos: %ld, %ld - stack: ", curIP->num, curIP->posx, curIP->posy);
		for(i = 0; i < curIP->SP; i++) fprintf(logfile, "%ld ", curIP->stack[i]);
		fprintf(logfile, "- command: %c\n", (int)c);
	}
	if(progmode == MODE_DEBUG) {
		fprintf(stderr, "execIP(IP:%ld): pos: %ld, %ld - stack: ", curIP->num, curIP->posx, curIP->posy);
		for(i = 0; i < curIP->SP; i++) fprintf(stderr, "%ld ", curIP->stack[i]);
		fprintf(stderr, "- command: %c\n", (int)c);
	}

	if((c >= '0') && (c <= '9')) { push(curIP, c - '0'); return r; }
	if((c >= 'a') && (c <= 'f')) { if(flags & FLAG_BEFUNGE93) return r; push(curIP, c - 87); return r; }
	if((c >= 'A') && (c <= 'Z')) {
		if(flags & FLAG_BEFUNGE93) return r;
		if(curIP->extfunctable[c - 'A']) {
			extfunc = curIP->extfunctable[c - 'A'];
			return (*extfunc)(curIP);
		} else {
			curIP->dx = -curIP->dx; curIP->dy = -curIP->dy;
			return r;
		}
	}
	
	/* switch, execute the command */
	switch(c) {
	case '>': /* set delta to (1, 0) */
		curIP->dx = 1; curIP->dy = 0;
		break;
	case '<': /* set delta to (-1, 0) */
		curIP->dx = -1; curIP->dy = 0;
		break;
	case '^': /* set delta to (0, -1) */
		curIP->dx = 0; curIP->dy = -1;
		break;
	case 'v': /* set delta to (0, 1) */
		curIP->dx = 0; curIP->dy = 1;
		break;
	case '?': /* set delta to random direction */
		if(rand()%2) { curIP->dx = (rand()%2)*2 - 1; curIP->dy = 0; }
		else { curIP->dx = 0; curIP->dy = (rand()%2)*2 - 1; }
		break;
	case ']': /* rotate delta 90 degrees right */
		if(flags & FLAG_BEFUNGE93) break;
		i = curIP->dx; curIP->dx = curIP->dy; curIP->dy = -i;
		break;
	case '[': /* rotate delta 90 degrees left */
		if(flags & FLAG_BEFUNGE93) break;
		i = curIP->dy; curIP->dy = curIP->dx; curIP->dx = -i;
		break;
	case ';': /* fly to next ';'. lame. too easy */
		if(flags & FLAG_BEFUNGE93) break;
		do {
			moveIP(curIP);
		} while(get(curIP->posx, curIP->posy) != ';');
		break;
	case 'k': /* iterate next instruction */
		if(flags & FLAG_BEFUNGE93) break;
		i = pop(curIP);
		moveIP(curIP);
		a = get(curIP->posx, curIP->posy);
		if((a != ' ') && (a != ';') && (a != '#')) for(b = 0; b < i; b++) execIP(NULL, a);
		break;
	case '(': /* load an extension */
		if(flags & FLAG_BEFUNGE93) break;
		a = pop(curIP);
		i = 0;
		for(b = 0; b < a; b++) {
			i <<= 8;
			i += pop(curIP);
		}
		/*if(!loadext(curIP, i)) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; }*/
		break;
	case ')': /* unload an extension */
		if(flags & FLAG_BEFUNGE93) break;
		/*if(!unloadext(curIP)) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; }*/
		break;
	case 'h': /* unimplemented; would go 'up'. now works like r */
	case 'l': /* unimplemented; would go 'down'. now works like r */
	case 'm': /* unimplemented; would be the up-down-if. now works like r */
	case '{': /* unimplemented; stack stack not implemented. now works like r */
	case '}': /* unimplemented; stack stack not implemented. now works like r */
	case 'u': /* unimplemented; stack stack not implemented. now works like r */
	case 'r': /* reverse delta */
		if(flags & FLAG_BEFUNGE93) break;
		curIP->dx = -curIP->dx; curIP->dy = -curIP->dy;
		break;
	case 'x': /* absolute vector: pops two values from stack and sets the delta so. a bit lame, IMO */
		if(flags & FLAG_BEFUNGE93) break;
		curIP->dy = pop(curIP); curIP->dx = pop(curIP);
		break;
	case '#': /* skip the next would-be cell. we do this with a additional call to moveIP, to get space-wrappings etc right */
		if(flags & FLAG_BEFUNGE93) moveIP93(curIP);
		else moveIP(curIP);
		break;
	case '@': /* kill IP */
		destroyIP(curIP);
		r = 0;
		break;
	case 'q': /* quits the program, with a return value popped */
		if(flags & FLAG_BEFUNGE93) break;
		terminate(pop(curIP));
		break;
	case 'j': /* jumps over arbitary number of spaces. this is _far_ too easy. I wouldn't use this one */
		if(flags & FLAG_BEFUNGE93) break;
		i = pop(curIP);
		if(i >= 0) while(i--) moveIP(curIP);
		else {
			i = -i; curIP->dx = -curIP->dx; curIP->dy = -curIP->dy;
			while(i--) moveIP(curIP);
			curIP->dx = -curIP->dx; curIP->dy = -curIP->dy;
		}
		break;
	case '!': /* logical not. pops a value and pushes a logical negation */
		push(curIP, !pop(curIP));
		break;
	case '`': /* greater-than. pops two values, then pushes 1 if the second one was greater. otherwise pushes 0 */
		push(curIP, pop(curIP) <= pop(curIP));
		break;
	case '_': /* > if zero, < if non-zero */
		curIP->dy = 0;
		if(pop(curIP)) curIP->dx = -1; else curIP->dx = 1;
		break;
	case '|': /* v if zero, ^ if non-zero */
		curIP->dx = 0;
		if(pop(curIP)) curIP->dy = -1; else curIP->dy = 1;
		break;
	case 'w': /* compare. [ if smaller, ] if greater, nop if equal */
		if(flags & FLAG_BEFUNGE93) break;
		b = pop(curIP); a = pop(curIP);
		if(a < b) { i = curIP->dy; curIP->dy = curIP->dx; curIP->dx = -i; }
		if(a > b) { i = curIP->dx; curIP->dx = curIP->dy; curIP->dy = -i; }
		break;
	case '+': /* add */
		push(curIP, pop(curIP) + pop(curIP));
		break;
	case '*': /* multiply */
		push(curIP, pop(curIP) * pop(curIP));
		break;
	case '-': /* substract */
		b = pop(curIP);
		push(curIP, pop(curIP) - b);
		break;
	case '/': /* divide */
		if(!(b = pop(curIP))) push(curIP, 0);
		else push(curIP, pop(curIP) / b);
		break;
	case '%': /* reminder */
		if(!(b = pop(curIP))) push(curIP, 0);
		else push(curIP, pop(curIP) % b);
		break;
	case '\"': /* toggle stringmode */
		curIP->stringmode = !curIP->stringmode;
		break;
	case '\'': /* fetch character */
		if(flags & FLAG_BEFUNGE93) break;
		moveIP(curIP); push(curIP, get(curIP->posx, curIP->posy));
		break;
	case '$': /* pop and discard */
		pop(curIP);
		break;
	case ':': /* duplicate. we won't pop and push twice, because push is a slow instruction */
		if(curIP->SP == 0) { push(curIP, 0); push(curIP, 0); }
		else { push(curIP, curIP->stack[curIP->SP - 1]); }
		break;
	case '\\': /* swap */
		a = pop(curIP); b = pop(curIP);
		push(curIP, a); push(curIP, b);
		break;
	case 'n': /* clear stack */
		if(flags & FLAG_BEFUNGE93) break;
		curIP->SP = 0;
		break;
	case 'g': /* get from funge-space */
		a = pop(curIP); b = pop(curIP);
		push(curIP, get(b, a));
		break;
	case 'p': /* put to funge-space */
		a = pop(curIP); b = pop(curIP);
		put(b, a, pop(curIP));
		break;
	case '.': /* output decimal */
		(curIP->outdecfunc)(pop(curIP));
		break;
	case ',': /* output character */
		(curIP->outchrfunc)(pop(curIP));
		break;
	case '&': /* input decimal */
		push(curIP, (curIP->inpdecfunc)());
		break;
	case '~': /* input character */
		push(curIP, (curIP->inpchrfunc)());
		break;
	case 'i': /* file input */
		if(flags & FLAG_BEFUNGE93) break;
		if(inpfile(curIP)) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; }
		break;
	case 'o': /* file output */
		if(flags & FLAG_BEFUNGE93) break;
		if(outfile(curIP)) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; }
		break;
	case '=': /* execute command */
		if(flags & FLAG_BEFUNGE93) break;
		push(curIP, execcmd(curIP));
		break;
	case 'y': /* get sysinfo */
		if(flags & FLAG_BEFUNGE93) break;
		systeminfo(curIP);
		break;
	case 't': /* split to two threads */
		if(flags & FLAG_BEFUNGE93) break;
		newIP = createIP(curIP);
		newIP->dx = -newIP->dx; newIP->dy = -newIP->dy;
		moveIP(newIP);
		break;
	default:
		if(flags & FLAG_BEFUNGE93) break;
		else curIP->dx = -curIP->dx; curIP->dy = -curIP->dy;
	}
	
	return r;
}

void moveIP(IP * curIP) {
	long int newx = 0; long int newy = 0;

	if(curIP->stringmode) {
		if((get(curIP->posx, curIP->posy) != ' ') && (get(curIP->posx + curIP->dx, curIP->posy + curIP->dy) == ' ')) {
			curIP->posx += curIP->dx; curIP->posy += curIP->dy;
			return;
		}
	}

	do {
		curIP->posx += curIP->dx; curIP->posy += curIP->dy;
		if((curIP->posx / 64 > blockmaxx) || (curIP->posx / 64 < blockminx) || (curIP->posy / 64 > blockmaxy) || (curIP->posy / 64 < blockminy)) {
			do {
				curIP->posx -= curIP->dx; curIP->posy -= curIP->dy;
				if(get(curIP->posx, curIP->posy) != ' ') { newx = curIP->posx; newy = curIP->posy; }
			} while ((curIP->posx / 64 <= blockmaxx) && (curIP->posx / 64 >= blockminx) && (curIP->posy / 64 <= blockmaxy) && (curIP->posy / 64 >= blockminy));
			curIP->posx = newx; curIP->posy = newy;
		}
	} while(get(curIP->posx, curIP->posy) == ' ');

	return;
}

void moveIP93(IP * curIP) {
	curIP->posx += curIP->dx; curIP->posy += curIP->dy;

	if(curIP->posx > pf_maxx) curIP->posx = pf_minx;
	if(curIP->posx < pf_minx) curIP->posx = pf_maxx;
	if(curIP->posy > pf_maxy) curIP->posy = pf_miny;
	if(curIP->posy < pf_miny) curIP->posy = pf_maxy;
	
	return;
}

long int pop(IP * curIP) {
	/* just return the value. note, that the stack will grow dynamically, but never get smaller. on underflow, return 0 */
	return curIP->SP ? curIP->stack[--curIP->SP] : 0;
}

void push(IP * curIP, long int data) {
	long int * tmpstack;

	if(flags & FLAG_STAT) {
		if(statinfo.minstacknum > data) statinfo.minstacknum = data;
		if(statinfo.maxstacknum < data) statinfo.maxstacknum = data;
		if(curIP->SP >= statinfo.max_stack) statinfo.max_stack = curIP->SP + 1;
	}

	if(curIP->SP >= curIP->stacksize) {
		/* allocate more stack space if necessary */
		if(flags & FLAG_DEBUG) fprintf(stderr, "push(IP:%ld, %ld) had to allocate more stack size (stack is %ld bytes now)\n", curIP->num, data, curIP->stacksize);
		curIP->stacksize += 1024;
		tmpstack = (long int *)malloc(sizeof(long int) * curIP->stacksize);
		if(!tmpstack) fail("malloc() failed while resizing stack");
		memcpy(tmpstack, curIP->stack, curIP->stacksize);
		free(curIP->stack);
		curIP->stack = tmpstack;
	}
	
	curIP->stack[curIP->SP] = data;
	curIP->SP++;
	
	return;
}

long int get(long int x, long int y) {
	unsigned long int i, j, k;

	if((x == lastgetx) && (y == lastgety)) return lastgetv;
	lastgetx = x; lastgety = y;

	if(flags & FLAG_LIMITED) if (((x < pf_minx) || (x > pf_maxx) || (y < pf_miny) || (y > pf_maxy))) return 32;

	if(!((x & (~(long int)0x3FF)) && (y & (~(long int)0x3FF)))) {
		i = x >> 6; j = y >> 6;
		if(sbquick[i][j]) return lastgetv=sbquick[i][j]->data[y & 0x3F][x & 0x3F];
		else return lastgetv=32;
	}
	
	/* search through the space block stack for the needed block */
	if(x >= 0) i = x >> 6;
	else { i = -1 - ((-x) / 64); x = (i * (-64)) + x; }
	if(y >= 0) j = y >> 6;
	else { j = -1 - ((-y) / 64); y = (j * (-64)) + y; }
	for(k = 0; k < SBcount; k++)
		if((SBs[k]->offsetx == i) && (SBs[k]->offsety == j)) return lastgetv=SBs[k]->data[y & 0x3F][x & 0x3F];
	/* return space, if asked for anything not allocated */
	return lastgetv=32;
}

void put(long int x, long int y, long int data) {
	unsigned long int i, j, k;
	spaceblock ** tmpSBs; spaceblock * newblock;
	
	if((x == lastgetx) && (y == lastgety)) lastgetv = data;

	if(flags & FLAG_LIMITED) if (((x < pf_minx) || (x > pf_maxx) || (y < pf_miny) || (y > pf_maxy))) return;

	if(!((x & (~(long int)0x3FF)) && (y & (~(long int)0x3FF)))) {
		i = x >> 6; j = y >> 6;
		if(sbquick[i][j]) { sbquick[i][j]->data[y & 0x3F][x & 0x3F] = data; return; }
		if(flags & FLAG_DEBUG) fprintf(stderr, "put(%ld, %ld, %ld)[quick] had to allocate a new block (#%ld)\n", x, y, data, SBcount);
		SBcount++;
		if(SBcount >= SBstacksize) {
			SBstacksize += 256;
			tmpSBs = (spaceblock **)malloc(sizeof(spaceblock *) * SBstacksize);
			if(!tmpSBs) fail("malloc() failed while allocating more fungespace");
			memcpy(tmpSBs, SBs, sizeof(spaceblock *) * (SBstacksize - 256));
			free(SBs);
			SBs = tmpSBs;
		}
		newblock = (spaceblock *)malloc(sizeof(spaceblock));
		if(!newblock) fail("malloc() failed for a new block of fungespace");
		newblock->offsetx = i; newblock->offsety = j;
		for(k = 0; k < 4096; k++) ((long int *)newblock->data)[k] = 32;
		SBs[SBcount - 1] = newblock;
		sbquick[i][j] = newblock;
		
		/* and the data, too */
		newblock->data[y & 0x3F][x & 0x3F] = data;

		/* and update blockmaxs and blockmins */
		for(k = 0; k < SBcount; k++) {
			if(blockmaxx < i) blockmaxx = i;
			if(blockminx > i) blockminx = i;
			if(blockmaxy < j) blockmaxy = j;
			if(blockminy > j) blockminy = j;
		}
	}
	
	/* search through the space block stack for the needed block */
	if(x >= 0) i = x >> 6;
	else { i = -1 - ((-x) / 64); x = (i * (-64)) + x; }
	if(y >= 0) j = y >> 6;
	else { j = -1 - ((-y) / 64); y = (j * (-64)) + y; }
	for(k = 0; k < SBcount; k++)
		if((SBs[k]->offsetx == i) && (SBs[k]->offsety == j)) { SBs[k]->data[y & 0x3F][x & 0x3F] = data; return; }

	/* if we try to put into a non-allocated area, we'll just have to allocate it */
	if(flags & FLAG_DEBUG) fprintf(stderr, "put(%ld, %ld, %ld) had to allocate a new block (#%ld)\n", x, y, data, SBcount);
	SBcount++;
	if(SBcount >= SBstacksize) {
		SBstacksize += 256;
		tmpSBs = (spaceblock **)malloc(sizeof(spaceblock *) * SBstacksize);
		if(!tmpSBs) fail("malloc() failed while allocating more fungespace");
		memcpy(tmpSBs, SBs, sizeof(spaceblock *) * (SBstacksize - 256));
		free(SBs);
		SBs = tmpSBs;
	}

	/* then we allocate the new block and add it to the space block stack */
	newblock = (spaceblock *)malloc(sizeof(spaceblock));
	if(!newblock) fail("malloc() failed for a new block of fungespace");
	newblock->offsetx = i; newblock->offsety = j;
	for(k = 0; k < 4096; k++) ((long int *)newblock->data)[k] = 32;
	SBs[SBcount - 1] = newblock;

	/* and the data, too */
	newblock->data[y & 0x3F][x & 0x3F] = data;

	/* and update blockmaxs and blockmins */
	for(k = 0; k < SBcount; k++) {
		if(blockmaxx < i) blockmaxx = i;
		if(blockminx > i) blockminx = i;
		if(blockmaxy < j) blockmaxy = j;
		if(blockminy > j) blockminy = j;
	}

	return;
}

void systeminfo(IP * curIP) {
	long int req;
	int i, j;
	struct tm * timest; time_t timev;
	
	req = pop(curIP);
	if(req <= 0) req = 0;
	
	timev = time(NULL); timest = gmtime(&timev);
	
	if(!req || req==20) {
		push(curIP, 0);
		i = 0; while(*environ[i])
			for(j = strlen(environ[i]) + 1; j >= 0; j--) push(curIP, environ[i][j]);
	}
	if(!req || req==19) {
		push(curIP, 0);
		for(i = cmdargc; i >= 0; i--)
			for(j = strlen(cmdargv[i]) + 1; j >= 0; j--) push(curIP, cmdargv[i][j]);
	}
	if(!req || req==18) {
		push(curIP, curIP->SP);
	}
	if(!req || req==17) {
		push(curIP, 1);
	}
	if(!req || req==16) {
		push(curIP, (timest->tm_hour << 16) + (timest->tm_min << 8) + timest->tm_sec);
	}
	if(!req || req==15) {
		push(curIP, (timest->tm_year << 16) + (timest->tm_mon << 8) + timest->tm_mday - 1);
	}
	if(!req || req==14) {
		push(curIP, (blockmaxx<<6) + 63);
		push(curIP, (blockmaxy<<6) + 63);
	}
	if(!req || req==13) {
		push(curIP, blockminx<<6);
		push(curIP, blockminy<<6);
	}
	if(!req || req==12) {
		push(curIP, curIP->soffx); push(curIP, curIP->soffy);
	}
	if(!req || req==11) {
		push(curIP, curIP->dx); push(curIP, curIP->dy);
	}
	if(!req || req==10) {
		push(curIP, curIP->posx); push(curIP, curIP->posy);
	}
	if(!req || req==9) {
		push(curIP, 0);
	}
	if(!req || req==8) {
		push(curIP, curIP->num);
	}
	if(!req || req==7) {
		push(curIP, 2);
	}
	if(!req || req==6) {
		push(curIP, '/');
	}
	if(!req || req==5) {
		push(curIP, 1);
	}
	if(!req || req==4) {
		push(curIP, version_num);
	}
	if(!req || req==3) {
		push(curIP, ('G' << 24) + ('L' << 16) + ('F' << 8) + 'U');
	}
	if(!req || req==2) {
		push(curIP, sizeof(long int));
	}
	if(!req || req==1) {
		push(curIP, 0x1F);
	}
}

char * logtime(void) {
	time_t timev; struct tm * timegm;
	
	if(!timestr) timestr = (char *)malloc(20);
	timev = time(NULL);
	timegm = gmtime(&timev);
	sprintf(timestr, "%02d.%02d.%04d %02d:%02d:%02d", timegm->tm_mday, timegm->tm_mon + 1, timegm->tm_year + 1900, timegm->tm_hour + 1, timegm->tm_min + 1, timegm->tm_sec + 1);
	return timestr;
}

void logclose(void) {
	fprintf(logfile, "[%s] --- GLfunge98-%s log endns ---\n", logtime(), version_str);
	fclose(logfile);
	free(timestr); timestr = NULL;
	return;
}

#ifndef WIN32
#ifndef __hpux
int loadext(IP * curIP, long int fprint) {
	char libname[256];
	void * libhandle;
	int i, j;
	void * (*libfunc2)(void *, int, int *);
	int (*initfunc)(void);
	void * tmpfunc;
	char * errorstr;
	struct s_extsave * newsave;

	sprintf(libname, "%s/GLfunge98_ext_%c%c%c%c.so", EXTDIR, (int)((fprint & 0xff000000) >> 24), (int)((fprint & 0xff0000) >> 16), (int)((fprint & 0xff00) >> 8), (int)(fprint & 0xff));
	libhandle = dlopen(libname, RTLD_LAZY);
	if(!libhandle) {
		fputs(dlerror(), stderr);
		fputc('\n', stderr);
		return 0;
	}

	initfunc = dlsym(libhandle, "initext");
	errorstr = dlerror();
	if(!errorstr) if(!(*initfunc)()) return 0;

	newsave = (struct s_extsave *)malloc(sizeof(struct s_extsave));
	newsave->prevext = curIP->prevlib;
	newsave->extlibhandle = curIP->curlib;
	curIP->prevlib = newsave;
	curIP->curlib = libhandle;

	libfunc2 = dlsym(libhandle, "registercmd");
	errorstr = dlerror();
	if(errorstr) {
		fputs(errorstr, stderr);
		fputc('\n', stderr);
		return 0;
	}

	for(i = 'A'; i <= 'Z'; i++) {
		tmpfunc = (*libfunc2)(curIP->extfunctable[i - 'A'], i, &j);
		if(j) curIP->extfunctable[i - 'A'] = tmpfunc;
	}

	return 1;
}

int unloadext(IP * curIP) {
	void * (*libfunc2)(int, int *);
	int (*deinitfunc)(void);
	void * libhandle = NULL;
	char * errorstr;
	int i, j;
	void * tmpfunc;

	if(!curIP->curlib) return 0;

	libhandle = curIP->curlib;

	curIP->curlib = curIP->prevlib->extlibhandle;
	curIP->prevlib = curIP->prevlib->prevext;

	libfunc2 = dlsym(libhandle, "restorecmd");
	errorstr = dlerror();
	if(errorstr) {
		fputs(errorstr, stderr);
		fputc('\n', stderr);
		return 0;
	}

	for(i = 'A'; i <= 'Z'; i++) {
		tmpfunc = (*libfunc2)(i, &j);
		if(j) curIP->extfunctable[i - 'A'] = tmpfunc;
	}

	deinitfunc = dlsym(libhandle, "deinitext");
	errorstr = dlerror();
	if(!errorstr) if(!(*deinitfunc)()) return 0;

	dlclose(libhandle);

	return 1;
}
#endif /* !WIN32 */
#endif
