/*
	GLfunge98 'FPSP' extension (fingerprint 0x46505350)
	Implementation Copyright (C) 2000, Heikki Kallasjoki
	Semantics by Mike Riley's FPSP extension.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/


#include <stdlib.h>
#include <IPstruct.h>
#include <math.h>
#include <stdio.h>
#include <string.h>


/* actual functions */
void * oldfunks[21];
int cmd_A(IP * curIP) { /* "Add" */
	float a, b, n; long int ai, bi, ni;
	bi = (*curIP->popfunc)(curIP);
	ai = (*curIP->popfunc)(curIP);
	a = *((float *)&ai);
	b = *((float *)&bi);
	n = a + b;
	ni = *((long int *)&n);
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}
int cmd_B(IP * curIP) { /* "Sine" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = sin(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_C(IP * curIP) { /* "Cosine" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = cos(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_D(IP * curIP) { /* "Divide" */
	float a, b, n; long int ai, bi, ni;
	bi = (*curIP->popfunc)(curIP);
	ai = (*curIP->popfunc)(curIP);
	a = *((float *)&ai);
	b = *((float *)&bi);
	if(b == 0) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	n = a / b;
	ni = *((long int *)&n);
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}
int cmd_E(IP * curIP) { /* "Arc sine" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = asin(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_F(IP * curIP) { /* "Convert integer to floating point" */
	long int ni; float n;
	n = (*curIP->popfunc)(curIP);
	ni = *((long int *)&n);
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}
int cmd_G(IP * curIP) { /* "Arc tangent" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = atan(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_H(IP * curIP) { /* "Arc cosine" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = acos(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_I(IP * curIP) { /* "Convert floating point to integer" */
	long int ni; float n;
	ni = (*curIP->popfunc)(curIP);
	n = *((float *)&ni);
	ni = n;
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}
int cmd_K(IP * curIP) { /* "Natural logarithm" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = log(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_L(IP * curIP) { /* "Base-10 logarithm" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = log10(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_M(IP * curIP) { /* "Multiply" */
	float a, b, n; long int ai, bi, ni;
	bi = (*curIP->popfunc)(curIP);
	ai = (*curIP->popfunc)(curIP);
	a = *((float *)&ai);
	b = *((float *)&bi);
	n = a * b;
	ni = *((long int *)&n);
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}
int cmd_N(IP * curIP) {
	return 1;
}
int cmd_P(IP * curIP) { /* "Print" */
	long int ni, ni2; float n; char buff[128];
	ni = (*curIP->popfunc)(curIP);
	n = *((float *)&ni);
	sprintf(buff, "%f", n);
	ni2 = strlen(buff) + 1;
	for(ni = 0; ni < ni2; ni++) (*curIP->outchrfunc)(buff[ni]);
	(*curIP->outchrfunc)(' ');
	return 1;
}
int cmd_Q(IP * curIP) { /* "Square root" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = sqrt(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_R(IP * curIP) { /* "Convert ascii to floating point" */
	float n; long int ni, len; char * buff, * buff2;
	len = 256; buff = (char *)malloc(len);
	if(!buff) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	ni = 0;
	while((buff[ni++] = (*curIP->popfunc)(curIP)) != 0) if(ni == len) {
			len += 256; buff2 = (char *)malloc(len);
			if(!buff2) { free(buff); curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
			memcpy(buff2, buff, len - 256);
			free(buff); buff = buff2;
		}
	n = strtod(buff, (char **)NULL);
	ni = *((long int *)&n);
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}
int cmd_S(IP * curIP) { /* "Substract" */
	float a, b, n; long int ai, bi, ni;
	bi = (*curIP->popfunc)(curIP);
	ai = (*curIP->popfunc)(curIP);
	a = *((float *)&ai);
	b = *((float *)&bi);
	n = a - b;
	ni = *((long int *)&n);
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}
int cmd_T(IP * curIP) { /* "Tangent" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = tan(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_V(IP * curIP) { /* "Absolute value" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = abs(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_X(IP * curIP) { /* "Exponential" */
	float n1, n2; long int n1i, n2i;
	n1i = (*curIP->popfunc)(curIP);
	n1 = *((float *)&n1i);
	n2 = exp(n1);
	n2i = *((long int *)&n2);
	(*curIP->pushfunc)(curIP, n2i);
	return 1;
}
int cmd_Y(IP * curIP) { /* "x to the power of y" */
	float a, b, n; long int ai, bi, ni;
	bi = (*curIP->popfunc)(curIP);
	ai = (*curIP->popfunc)(curIP);
	a = *((float *)&ai);
	b = *((float *)&bi);
	n = pow(a, b);
	ni = *((long int *)&n);
	(*curIP->pushfunc)(curIP, ni);
	return 1;
}

/* necessary functions for GLfunge98-extension */

/* registercmd - registers a command function to a funge-command */
void * registercmd(void * oldfunc, int cmdchar, int * statusval) {
	*statusval = 1;

	switch(cmdchar) {
	case 'A':
		oldfunks[0] = oldfunc;
		return cmd_A;
	case 'B':
		oldfunks[1] = oldfunc;
		return cmd_B;
	case 'C':
		oldfunks[2] = oldfunc;
		return cmd_C;
	case 'D':
		oldfunks[3] = oldfunc;
		return cmd_D;
	case 'E':
		oldfunks[4] = oldfunc;
		return cmd_E;
	case 'F':
		oldfunks[5] = oldfunc;
		return cmd_F;
	case 'G':
		oldfunks[6] = oldfunc;
		return cmd_G;
	case 'H':
		oldfunks[7] = oldfunc;
		return cmd_H;
	case 'I':
		oldfunks[8] = oldfunc;
		return cmd_I;
	case 'K':
		oldfunks[9] = oldfunc;
		return cmd_K;
	case 'L':
		oldfunks[10] = oldfunc;
		return cmd_L;
	case 'M':
		oldfunks[11] = oldfunc;
		return cmd_M;
	case 'N':
		oldfunks[12] = oldfunc;
		return cmd_N;
	case 'P':
		oldfunks[13] = oldfunc;
		return cmd_P;
	case 'Q':
		oldfunks[14] = oldfunc;
		return cmd_Q;
	case 'R':
		oldfunks[15] = oldfunc;
		return cmd_R;
	case 'S':
		oldfunks[16] = oldfunc;
		return cmd_S;
	case 'T':
		oldfunks[17] = oldfunc;
		return cmd_T;
	case 'V':
		oldfunks[18] = oldfunc;
		return cmd_V;
	case 'X':
		oldfunks[19] = oldfunc;
		return cmd_X;
	case 'Y':
		oldfunks[20] = oldfunc;
		return cmd_Y;
	}

	*statusval = 0;
	return NULL;
}

/* restorecmd - restores old commands */
void * restorecmd(int cmdchar, int * statusval) {
	*statusval = 1;

	switch(cmdchar) {
	case 'A':
		return oldfunks[0];
	case 'B':
		return oldfunks[1];
	case 'C':
		return oldfunks[2];
	case 'D':
		return oldfunks[3];
	case 'E':
		return oldfunks[4];
	case 'F':
		return oldfunks[5];
	case 'G':
		return oldfunks[6];
	case 'H':
		return oldfunks[7];
	case 'I':
		return oldfunks[8];
	case 'K':
		return oldfunks[9];
	case 'L':
		return oldfunks[10];
	case 'M':
		return oldfunks[11];
	case 'N':
		return oldfunks[12];
	case 'P':
		return oldfunks[13];
	case 'Q':
		return oldfunks[14];
	case 'R':
		return oldfunks[15];
	case 'S':
		return oldfunks[16];
	case 'T':
		return oldfunks[17];
	case 'V':
		return oldfunks[18];
	case 'X':
		return oldfunks[19];
	case 'Y':
		return oldfunks[20];
	}

	*statusval = 0;
	return NULL;
}


