/*
	GLfunge98 'SOCK' extension, fingerprint 0x534F434b
	Copyright (C) 2000, Heikki Kallasjoki
	Semantics as Mike Riley's SOCK-extension
	Additions by Heikki Kallasjoki, in SCKE-extension

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/


#include <stdlib.h>
#include <IPstruct.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <arpa/inet.h>


/* global variables */
/*void * prevlib;*/


/* actual functions */
void * oldfunks[11];
int cmd_A(IP * curIP) { /* "Accept connection" */
	struct sockaddr_in newaddr;
	long int i, s; int k;

	s = (*curIP->popfunc)(curIP);

	k = sizeof(struct sockaddr_in);
	i = accept(s, (struct sockaddr *)&newaddr, &k);
	if(i == -1) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	(*curIP->pushfunc)(curIP, ntohs(newaddr.sin_port));
	(*curIP->pushfunc)(curIP, ntohl(newaddr.sin_addr.s_addr));
	(*curIP->pushfunc)(curIP, i);

	return 1;
}
int cmd_B(IP * curIP) { /* "Bind a socket */
	long int s, ct, prt, addr, i;
	struct sockaddr_in myaddr;

	addr = (*curIP->popfunc)(curIP);
	prt = (*curIP->popfunc)(curIP);
	ct = (*curIP->popfunc)(curIP);
	s = (*curIP->popfunc)(curIP);

	if(ct != 2) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	myaddr.sin_addr.s_addr = htonl(addr);
	myaddr.sin_port = htons(prt);
	myaddr.sin_family = AF_INET;

	i = bind(s, &myaddr, sizeof(struct sockaddr_in));
	if(i) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	return 1;
}
int cmd_C(IP * curIP) { /* "Open a connection" */
	long int s, ct, prt, addr, i;
	struct sockaddr_in myaddr;

	addr = (*curIP->popfunc)(curIP);
	prt = (*curIP->popfunc)(curIP);
	ct = (*curIP->popfunc)(curIP);
	s = (*curIP->popfunc)(curIP);

	if(ct != 2) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	
	myaddr.sin_addr.s_addr = htonl(addr);
	myaddr.sin_port = htons(prt);
	myaddr.sin_family = AF_INET;
	
	i = connect(s, &myaddr, sizeof(struct sockaddr_in));
	if(i) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	return 1;
}
int cmd_K(IP * curIP) { /* "Kill a connection" */
	long int mysock;
	
	mysock = (*curIP->popfunc)(curIP);
	
	close(mysock);
	
	return 1;
}
int cmd_L(IP * curIP) { /* "Set socket to listen mode" */
	long int mysock, mybacklog;
	
	mysock = (*curIP->popfunc)(curIP);
	mybacklog = (*curIP->popfunc)(curIP);

	if(listen(mysock, mybacklog) == -1) {
		curIP->dx = -curIP->dx; curIP->dy = -curIP->dy;
		return 1;
	}

	return 1;
}
int cmd_O(IP * curIP) {
	long int n, o, s, i, i1, i2;

	s = (*curIP->popfunc)(curIP);
	o = (*curIP->popfunc)(curIP);
	n = (*curIP->popfunc)(curIP);

	switch(o) {
	case 1: i = SO_DEBUG; break;
	case 2: i = SO_REUSEADDR; break;
	case 3: i = SO_KEEPALIVE; break;
	case 4: i = SO_DONTROUTE; break;
	case 5: i = SO_BROADCAST; break;
	case 6: i = SO_OOBINLINE; break;
	case 7:
		i1 = fcntl(s, F_GETFL);
		if(n) i2 = fcntl(s, F_SETFL, i1 & (!O_NONBLOCK));
		else i2 = fcntl(s, F_SETFL, i1 | O_NONBLOCK);
		if(i2 == -1) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
		return 1;
	default: curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1;
	}

	if(setsockopt(s, SOL_SOCKET, i, &n, sizeof(long int))) {
		curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1;
	}

	return 1;
}
int cmd_R(IP * curIP) { /* "Read from a socket" */
	long int s, l, v_x, v_y, i; unsigned char * buff;
	
	s = (*curIP->popfunc)(curIP);
	l = (*curIP->popfunc)(curIP);
	v_y = (*curIP->popfunc)(curIP);
	v_x = (*curIP->popfunc)(curIP);

	buff = (unsigned char *)malloc(l);
	if(!buff) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	i = recv(s, buff, l, 0);
	if(i == -1) { free(buff); curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	(*curIP->pushfunc)(curIP, i);
	
	for(i = 0; i < l; i++)
		(*curIP->putfunc)(v_x + i, v_y, buff[i]);
	
	free(buff);
	
	return 1;
}
int cmd_S(IP * curIP) { /* "Create a socket" */
	long int proto, family, type, i, i1, i2, i3;
	struct protoent * mypre = NULL;

	i1 = (*curIP->popfunc)(curIP);
	i2 = (*curIP->popfunc)(curIP);
	i3 = (*curIP->popfunc)(curIP);
	
	if(i1 == 1) mypre = getprotobyname("tcp");
	if(i1 == 2) mypre = getprotobyname("udp");
	if(!mypre) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	proto = mypre->p_proto;
	
	switch(i2) {
	case 1:
		type = SOCK_DGRAM;
		break;
	case 2:
		type = SOCK_STREAM;
		break;
	default:
		curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1;
	}
	
	switch(i3) {
	case 1:
		family = PF_UNIX;
		break;
	case 2:
		family = PF_INET;
		break;
	default:
		curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1;
	}

	i = socket(family, type, proto);
	if(i == -1) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	i1 = fcntl(i, F_GETFL);
	i2 = fcntl(i, F_SETFL, i1 | O_NONBLOCK);
	if(i2 == -1) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	(*curIP->pushfunc)(curIP, i);
	
	return 1;
}
int cmd_W(IP * curIP) { /* "Write to a socket" */
	long int s, l, v_x, v_y, i; unsigned char * buff;
	
	s = (*curIP->popfunc)(curIP);
	l = (*curIP->popfunc)(curIP);
	v_y = (*curIP->popfunc)(curIP);
	v_x = (*curIP->popfunc)(curIP);

	buff = (unsigned char *)malloc(l);
	if(!buff) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	
	for(i = 0; i < l; i++)
		buff[i] = (*curIP->getfunc)(v_x + i, v_y);

	i = send(s, buff, l, 0);
	if(i == -1) { free(buff); curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	
	(*curIP->pushfunc)(curIP, i);
	free(buff);

	return 1;
}
int cmd_H(IP * curIP) { /* "Get address by hostname" */
	char * str1, * str2;
	int lens, pos; long int mychr;
	struct in_addr myaddr;
	struct hostent * myhost;

	lens = 256;
	str1 = (char *)malloc(lens);
	if(!str1) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	pos = 0;
	while((mychr = (*curIP->popfunc)(curIP)) != 0) {
		str1[pos++] = mychr & 0xFF;
		if(pos == lens) {
			lens += 256;
			str2 = (char *)malloc(lens);
			if(!str2) { free(str1); curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
			memcpy(str2, str1, lens - 256);
			free(str1);
			str1 = str2;
		}
	}
	str1[pos] = '\0';

	if(inet_aton(str1, &myaddr)) {
		(*curIP->pushfunc)(curIP, ntohl(myaddr.s_addr));
		return 1;
	}
	
	myhost = gethostbyname(str1);
	if((!myhost) || (myhost->h_addrtype != AF_INET)) { free(str1); curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }

	(*curIP->pushfunc)(curIP, ntohl(*myhost->h_addr));
	
	return 1;
}
int cmd_P(IP * curIP) { /* "Peek for incoming data" */
	fd_set rfds; long int s; struct timeval tv; int i;

	s = (*curIP->popfunc)(curIP);
	
	FD_ZERO(&rfds);
	FD_SET(s, &rfds);
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	
	i = select(s + 1, &rfds, NULL, NULL, &tv);
	if(i == -1) { curIP->dx = -curIP->dx; curIP->dy = -curIP->dy; return 1; }
	
	(*curIP->pushfunc)(curIP, i);
	return 1;
}

/* necessary functions for GLfunge98-extension */

int initext(void) {
	return 1;
}

int deinitext(void) {
	return 1;
}

/* saveprev, restoreprev - saves and restores previous library */
/*void saveprev(void * plib) { prevlib = plib; }
void * restoreprev(void) { return prevlib; }*/

/* registercmd - registers a command function to a funge-command */
void * registercmd(void * oldfunc, int cmdchar, int * statusval) {
	*statusval = 1;

	switch(cmdchar) {
	case 'A':
		oldfunks[0] = oldfunc;
		return cmd_A;
	case 'B':
		oldfunks[1] = oldfunc;
		return cmd_B;
	case 'C':
		oldfunks[2] = oldfunc;
		return cmd_C;
	case 'K':
		oldfunks[3] = oldfunc;
		return cmd_K;
	case 'L':
		oldfunks[4] = oldfunc;
		return cmd_L;
	case 'O':
		oldfunks[5] = oldfunc;
		return cmd_O;
	case 'R':
		oldfunks[6] = oldfunc;
		return cmd_R;
	case 'S':
		oldfunks[7] = oldfunc;
		return cmd_S;
	case 'W':
		oldfunks[8] = oldfunc;
		return cmd_W;
	case 'H':
		oldfunks[9] = oldfunc;
		return cmd_H;
	case 'P':
		oldfunks[10] = oldfunc;
		return cmd_P;
	}

	*statusval = 0;
	return NULL;
}

/* restorecmd - restores old commands */
void * restorecmd(int cmdchar, int * statusval) {
	*statusval = 1;

	switch(cmdchar) {
	case 'A':
		return oldfunks[0];
	case 'B':
		return oldfunks[1];
	case 'C':
		return oldfunks[2];
	case 'K':
		return oldfunks[3];
	case 'L':
		return oldfunks[4];
	case 'O':
		return oldfunks[5];
	case 'R':
		return oldfunks[6];
	case 'S':
		return oldfunks[7];
	case 'W':
		return oldfunks[8];
	case 'H':
		return oldfunks[9];
	case 'P':
		return oldfunks[10];
	}

	*statusval = 0;
	return NULL;
}


