/*
	GLfunge98 - out_gl.c - function implementations for OpenGL output
	OpenGL-based implementation of Funge98
	Copyright (C) 2000, Heikki Kallasjoki

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA	
*/

#define GL_WINDOWX 640
#define GL_WINDOWY 480
#define GL_WINDOWTEXTX 80
#define GL_WINDOWTEXTY 25

#ifndef WIN32

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>

extern const char * version_str;
extern const long int version_num;

unsigned long int framecount_GL = 0;
Display * myDisplay;
int myScreenNum;
Window myWindow;
int myError = 0;
GLXContext myContext;
int fontlists = 0;
int printpos_x, printpos_y;
float printrastery[GL_WINDOWTEXTY];
char printdata[GL_WINDOWTEXTY][GL_WINDOWTEXTX];

void mySignalHandler(int signum) {
	framecount_GL++;
	signal(SIGVTALRM, mySignalHandler);
	return;
}

int (*oldErrorHandler)(Display *, XErrorEvent *);
int myErrorHandler(Display * thisDisplay, XErrorEvent * thisError) {
	myError = 1;
	return 0;
}

void draw_GL(void) {
	int i;

	glClear(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	glPushAttrib(GL_LIST_BIT);
	glListBase(fontlists);
	for(i = 0; i < GL_WINDOWTEXTY; i++) {
		glRasterPos2f(-1.0f, printrastery[i]);
		glCallLists(GL_WINDOWTEXTX, GL_UNSIGNED_BYTE, printdata[i]);
	}
	glPopAttrib();
	
	glFlush();

	glXSwapBuffers(myDisplay, myWindow);
}

int init_GL(void) {
	char * strtmp;
	XVisualInfo * myVisual;
	FILE * fontfile;
	char * fontstr;
	unsigned char * fontdata;
	int i;
	int thisVisAttributes[] = {GLX_USE_GL, GLX_RGBA, GLX_DOUBLEBUFFER, GLX_RED_SIZE, 5, GLX_GREEN_SIZE, 5, GLX_BLUE_SIZE, 5, GLX_ALPHA_SIZE, 1, None};

	fontstr = (char *)malloc(256);
	fontdata = (unsigned char *)malloc(256 * 8 * 16 * 4);
	if(!fontstr || !fontdata) return 1;

	/* open connection to X server */
	strtmp = getenv("DISPLAY");
	if(strtmp) myDisplay = XOpenDisplay(strtmp);
	else myDisplay = XOpenDisplay(":0.0");
	if(!myDisplay) return 1;
	myScreenNum = DefaultScreen(myDisplay);

	oldErrorHandler = XSetErrorHandler(myErrorHandler);

	/* create our window */
	myWindow = XCreateSimpleWindow(
		myDisplay,                                        /* use our display */
		RootWindow(myDisplay, myScreenNum),               /* root window is our parent */
		0, 0,                                             /* position, (0, 0) */
		GL_WINDOWX, GL_WINDOWY,                           /* our predefined window size */
		0, 0, 0);                                         /* borders and background are insignificant */
	if(myError) {
		XCloseDisplay(myDisplay);
		return 2;
	}
	sprintf(fontstr, "GLfunge98");
	XStoreName(myDisplay, myWindow, fontstr);

	/* map the window */
	XMapRaised(myDisplay, myWindow);
	if(myError) {
		XDestroyWindow(myDisplay, myWindow);
		XCloseDisplay(myDisplay);
		return 3;
	}
	
	/* create GLX drawing context */
	myVisual = glXChooseVisual(myDisplay, myScreenNum, thisVisAttributes);
	if(!myVisual) {
		XDestroyWindow(myDisplay, myWindow);
		XCloseDisplay(myDisplay);
		return 4;
	}
	myContext = glXCreateContext(myDisplay, myVisual, NULL, GL_TRUE);
	if(!myContext) {
		XFree(myVisual);
		XDestroyWindow(myDisplay, myWindow);
		XCloseDisplay(myDisplay);
		return 5;
	}
	
	/* attach the GLX context to our window */
	if(!glXMakeCurrent(myDisplay, myWindow, myContext)) {
		glXDestroyContext(myDisplay, myContext);
		XFree(myVisual);
		XDestroyWindow(myDisplay, myWindow);
		XCloseDisplay(myDisplay);
		return 6;
	}

	/* create basic text drawing font */
	sprintf(fontstr, "%s/data/font.tex", EXTDIR);
	fontfile = fopen(fontstr, "r");
	if(!fontfile) {
		free(fontstr); free(fontdata);
		glXDestroyContext(myDisplay, myContext);
		XFree(myVisual);
		XDestroyWindow(myDisplay, myWindow);
		XCloseDisplay(myDisplay);
		return 8;
	}
	fread(fontdata, 1, 256 * 8 * 16 * 4, fontfile);
	fontlists = glGenLists(256);
	if(!fontlists) {
		free(fontstr); free(fontdata); fclose(fontfile);
		glXDestroyContext(myDisplay, myContext);
		XFree(myVisual);
		XDestroyWindow(myDisplay, myWindow);
		XCloseDisplay(myDisplay);
		return 9;
	}
	for(i = 0; i < 256; i++) {
		glNewList(fontlists + i, GL_COMPILE);
			glDrawPixels(8, 16, GL_RGBA, GL_UNSIGNED_BYTE, fontdata + i * 8 * 16 * 4);
			glBitmap(0, 0, 0, 0, 8.0f, 0.0f, NULL);
		glEndList();
	}
	for(i = 0; i < GL_WINDOWTEXTY; i++) printrastery[i] = 1.0f - (((float)i + 1) / ((float)GL_WINDOWTEXTY)) * 2.0f;

	/* register our signal handler */
	signal(SIGVTALRM, mySignalHandler);

	/* set some GL state machine variables */
	glClearColor(0.0f, 1.0f, 0.0f, 0.0f);

	/* reset the 'cursor' to top of the screen */
	printpos_x = 0; printpos_y = 0;
	memset(printdata, 32, GL_WINDOWTEXTY * GL_WINDOWTEXTX);

	XFree(myVisual); free(fontstr); free(fontdata); fclose(fontfile);

	return 0;
}

void uninit_GL(void) {
	struct itimerval tmrv;

	/* kill our timer */
	tmrv.it_interval.tv_sec = 0;
	tmrv.it_interval.tv_usec = 0;
	tmrv.it_value.tv_sec = 0;
	tmrv.it_value.tv_usec = 0;
	setitimer(ITIMER_VIRTUAL, &tmrv, NULL);
	signal(SIGALRM, SIG_DFL);

	memcpy(printdata[11] + 25, "###############################", 31);
	memcpy(printdata[12] + 25, "# Program execution completed #", 31);
	memcpy(printdata[13] + 25, "###############################", 31);

	draw_GL();

	sleep(5);

	if(fontlists) glDeleteLists(fontlists, 256);
	glXDestroyContext(myDisplay, myContext);
	XDestroyWindow(myDisplay, myWindow);
	XCloseDisplay(myDisplay);

	return;
}

void outchr_GL(long int x) {
	if(x == '\x08') {
		if(printpos_x) printpos_x--;
		else {
			printpos_x = GL_WINDOWTEXTX - 1;
			if(printpos_y) printpos_y--;
			else printpos_y = GL_WINDOWTEXTY - 1;
		}
		printdata[printpos_y][printpos_x] = ' ';
		return;
	}

	if(x == 10) {
		printpos_x = 0;
		printpos_y++;
		if(printpos_y == GL_WINDOWTEXTY) {
			printpos_y--;
			memmove(printdata[0], printdata[1], (GL_WINDOWTEXTY - 1) * GL_WINDOWTEXTX);
			memset(printdata[GL_WINDOWTEXTY - 1], 32, GL_WINDOWTEXTX);
		}
		return;
	}

	if((x < 32) || (x > 255)) x = 32;

	printdata[printpos_y][printpos_x] = x & 0xff;
	if(printpos_x != GL_WINDOWTEXTX - 1) printpos_x++;
	else {
		printpos_x = 0;
		printpos_y++;
		if(printpos_y == GL_WINDOWTEXTY) {
			printpos_y--;
			memmove(printdata[0], printdata[1], (GL_WINDOWTEXTY - 1) * GL_WINDOWTEXTX);
			memset(printdata[GL_WINDOWTEXTY - 1], 32, GL_WINDOWTEXTX);
		}
	}

	return;
}

void outdec_GL(long int x) {
	char numval[32]; int i;
	
	sprintf(numval, "%ld ", x);
	for(i = 0; i <= strlen(numval); i++) outchr_GL(numval[i]);
	
	return;
}

long int inpchr_GL(void) { return 0; }
long int inpdec_GL(void) { return 0; }

#endif /* !WIN32 */
